package eu.dnetlib.repo.manager.service.aai.registry;

import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import eu.dnetlib.repo.manager.domain.dto.Role;
import eu.dnetlib.repo.manager.domain.dto.User;
import eu.dnetlib.repo.manager.service.aai.registry.utils.RegistryUtils;
import eu.dnetlib.repo.manager.utils.HttpUtils;
import org.apache.log4j.Logger;
import org.mitre.openid.connect.model.OIDCAuthenticationToken;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Service;

import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

@Service
public class RegistryCalls implements AaiRegistryService {

    private static final Logger logger = Logger.getLogger(RegistryCalls.class);

    private final String coid;
    public final HttpUtils httpUtils;
    public final RegistryUtils jsonUtils;

    @Autowired
    RegistryCalls(@Value("${services.provide.aai.registry.coid:2}") String coid,
                  HttpUtils httpUtils, RegistryUtils registryUtils) {
        this.coid = coid;
        this.httpUtils = httpUtils;
        this.jsonUtils = registryUtils;
    }

    private String mapType(String type, boolean communityMap) {
        if (type.equals("organization")) {
            type = "institution";
        } else if (type.equals("ri") && communityMap) {
            type = "community";
        }
        return type;
    }

    @Override
    public Integer getCoPersonIdByEmail() {
        try {
            OIDCAuthenticationToken authentication = (OIDCAuthenticationToken) SecurityContextHolder.getContext().getAuthentication();
            String email = authentication.getUserInfo().getEmail();
            Map<String, String> params = new HashMap<>();
            params.put("coid", coid);
            params.put("mail", email);
            JsonElement response = httpUtils.get("co_people.json", params);
            return (response != null) ? response.getAsJsonObject().get("CoPeople").getAsJsonArray().get(0).getAsJsonObject().get("Id").getAsInt() : null;
        } catch (Exception e) {
            logger.error("Get User info: An error occurred ", e);
            return null;
        }
    }

    @Override
    public Integer getCoPersonIdByEmail(String email) {
        Map<String, String> params = new HashMap<>();
        params.put("coid", coid);
        params.put("mail", email);
        JsonElement response = httpUtils.get("co_people.json", params);
        if (response != null) {
            JsonArray coPeople = response.getAsJsonObject().get("CoPeople").getAsJsonArray();
            if (coPeople.size() > 0) {
                return coPeople.get(0).getAsJsonObject().get("Id").getAsInt();
            }
        }
        return null;
    }

    @Override
    public List<Integer> getCoPersonIdsByEmail(String email) {
        List<Integer> coPersonIds = new ArrayList<>();
        Map<String, String> params = new HashMap<>();
        params.put("coid", coid);
        params.put("mail", email);
        JsonElement response = httpUtils.get("co_people.json", params);
        if (response != null) {
            JsonArray coPeople = response.getAsJsonObject().get("CoPeople").getAsJsonArray();
            for (int i = 0; i < coPeople.size(); i++) {
                coPersonIds.add(coPeople.get(i).getAsJsonObject().get("Id").getAsInt());
            }
        }
        return coPersonIds;
    }

    @Override
    public Integer getCoPersonIdByIdentifier() {
        try {
            OIDCAuthenticationToken authentication = (OIDCAuthenticationToken) SecurityContextHolder.getContext().getAuthentication();
            String sub = authentication.getUserInfo().getSub();
            Map<String, String> params = new HashMap<>();
            params.put("coid", coid);
            params.put("search.identifier", sub);
            JsonElement response = httpUtils.get("co_people.json", params);
            return (response != null) ? response.getAsJsonObject().get("CoPeople").getAsJsonArray().get(0).getAsJsonObject().get("Id").getAsInt() : null;
        } catch (Exception e) {
            logger.error("Get User info: An error occurred ", e);
            return null;
        }
    }

    public Integer getCoPersonIdByIdentifier(String sub) {
        Map<String, String> params = new HashMap<>();
        params.put("coid", coid);
        params.put("search.identifier", sub);
        JsonElement response = httpUtils.get("co_people.json", params);
        return (response != null) ? response.getAsJsonObject().get("CoPeople").getAsJsonArray().get(0).getAsJsonObject().get("Id").getAsInt() : null;
    }

    @Override
    public JsonArray getCous(String name) {
        Map<String, String> params = new HashMap<>();
        params.put("coid", coid);
        if (name != null) {
            params.put("name", URLEncoder.encode(name).toLowerCase());
        }
        JsonElement response = httpUtils.get("cous.json", params);
        return (response != null) ? response.getAsJsonObject().get("Cous").getAsJsonArray() : new JsonArray();
    }

    @Override
    public JsonArray getCous() {
        return getCous(null);
    }

    @Override
    public Integer getCouId(String name) {
        JsonArray cous = getCous(name);
        for (JsonElement cou : cous) {
            if (cou.getAsJsonObject().get("Name").getAsString().equalsIgnoreCase(name)) {
                return cou.getAsJsonObject().get("Id").getAsInt();
            }
        }
        return null;
    }

    @Override
    public Integer getCouId(String type, String id, boolean communityMap) {
        return getCouId(mapType(type, communityMap) + "." + id);
    }

    @Override
    public Integer getCouId(String type, String id) {
        return getCouId(type, id, true);
    }

    @Override
    public JsonArray getRoles(Integer coPersonId) {
        Map<String, String> params = new HashMap<>();
        params.put("copersonid", coPersonId.toString());
        JsonElement response = httpUtils.get("co_person_roles.json", params);
        return (response != null) ? response.getAsJsonObject().get("CoPersonRoles").getAsJsonArray() : new JsonArray();
    }

    @Override
    public JsonArray getRolesWithStatus(Integer coPersonId, RoleStatus status) {
        JsonArray roles = getRoles(coPersonId);
        if (status == null) {
            return roles;
        }
        JsonArray activeRoles = new JsonArray();
        for (JsonElement role : roles) {
            if (role.getAsJsonObject().get("Status").getAsString().equalsIgnoreCase(status.toString())) {
                activeRoles.add(role);
            }
        }
        return activeRoles;
    }

    @Override
    public Integer getRoleId(Integer coPersonId, Integer couId) {
        JsonArray roles = getRoles(coPersonId);
        for (JsonElement role : roles) {
            JsonObject object = role.getAsJsonObject();
            if (object.get("CouId").getAsInt() == couId && !object.get("Status").getAsString().equals("Deleted")) {
                return object.get("Id").getAsInt();
            }
        }
        return null;
    }

    @Override
    public JsonArray getUserGroups(Integer coPersonId) {
        Map<String, String> params = new HashMap<>();
        params.put("copersonid", coPersonId.toString());
        JsonElement response = httpUtils.get("co_groups.json", params);
        return (response != null) ? response.getAsJsonObject().get("CoGroups").getAsJsonArray() : new JsonArray();
    }

    @Override
    public JsonObject getUserAdminGroup(Integer coPersonId, Integer couId) {
        Map<String, String> params = new HashMap<>();
        params.put("copersonid", coPersonId.toString());
        JsonElement response = httpUtils.get("co_groups.json", params);
        JsonArray roles = (response != null) ? response.getAsJsonObject().get("CoGroups").getAsJsonArray() : new JsonArray();
        for (JsonElement role : roles) {
            JsonObject object = role.getAsJsonObject();
            if (object.get("CouId") != null && object.get("CouId").getAsInt() == couId) {
                if (object.get("Name").getAsString().contains("admins")) {
                    return object;
                }
            }
        }
        return null;
    }

    @Override
    public JsonArray getCouGroups(Integer couId) {
        Map<String, String> params = new HashMap<>();
        params.put("coid", coid);
        params.put("couid", couId.toString());
        JsonElement response = httpUtils.get("co_groups.json", params);
        return (response != null) ? response.getAsJsonObject().get("CoGroups").getAsJsonArray() : new JsonArray();
    }

    @Override
    public JsonObject getCouAdminGroup(Integer couId) {
        JsonArray groups = getCouGroups(couId);
        for (JsonElement group : groups) {
            if (group.getAsJsonObject().get("Name").getAsString().contains("admins")) {
                return group.getAsJsonObject();
            }
        }
        return null;
    }

    @Override
    public JsonArray getGroupMembers(Integer coGroupId) {
        Map<String, String> params = new HashMap<>();
        params.put("cogroupid", coGroupId.toString());
        JsonElement response = httpUtils.get("co_group_members.json", params);
        return (response != null) ? response.getAsJsonObject().get("CoGroupMembers").getAsJsonArray() : new JsonArray();
    }


    @Override
    public JsonArray getUserEmailByCouId(Integer couId, boolean admin) {
        Map<String, String> params = new HashMap<>();
        if (couId == null) {
            throw new IllegalArgumentException("Provided 'couId' is null");
        }
        params.put("couid", couId.toString());
        if (admin) {
            params.put("admin", "true");
        }
        JsonElement response = httpUtils.get("email_addresses.json", params);
        JsonArray infos = (response != null) ? response.getAsJsonObject().get("EmailAddresses").getAsJsonArray() : new JsonArray();
        JsonArray emails = new JsonArray();
        infos.forEach(info -> {
            JsonObject user = new JsonObject();
            boolean add = true;
            String email = info.getAsJsonObject().get("Mail").getAsString();
            for (JsonElement element : emails) {
                if (element.getAsJsonObject().get("email").getAsString().equals(email)) {
                    add = false;
                }
            }
            if (add) {
                user.addProperty("email", email);
                user.addProperty("memberSince", info.getAsJsonObject().get("Created").getAsString());
                emails.add(user);
            }
        });
        return emails;
    }

    @Override
    public JsonArray getUsersByCouId(Integer couId) {
        Map<String, String> params = new HashMap<>();
        params.put("couid", couId.toString());
        JsonElement response = httpUtils.get("co_person_roles.json", params);
        JsonArray infos = (response != null) ? response.getAsJsonObject().get("CoPersonRoles").getAsJsonArray() : new JsonArray();
//        JsonArray users = new JsonArray();
//        infos.forEach(info -> {
//            JsonObject user = new JsonObject();
//            user.addProperty("email", info.getAsJsonObject().get("Mail").getAsString());
//            user.addProperty("memberSince", info.getAsJsonObject().get("Created").getAsString());
//            emails.add(user);
//        });
        return infos;
    }

    @Override
    public List<User> getUsers(Integer couId) {
        List<User> users = new ArrayList<>();
        JsonArray infos = getUserEmailByCouId(couId, false);

        infos.forEach(info -> {
            User user = new User();
            // TODO: should add firstname and lastname and sub of user
            user.setEmail(info.getAsJsonObject().get("email").getAsString());

            users.add(user);
        });
        return users;
    }

    @Override
    public JsonArray getUserNamesByCouId(Integer couId, boolean admin) {
        Map<String, String> params = new HashMap<>();
        params.put("couid", couId.toString());
        if (admin) {
            params.put("admin", "true");
        }
        JsonElement response = httpUtils.get("names.json", params);
        JsonArray infos = (response != null) ? response.getAsJsonObject().get("Names").getAsJsonArray() : new JsonArray();
        JsonArray names = new JsonArray();
        infos.forEach(info -> {
            JsonObject user = new JsonObject();
            user.addProperty("name", info.getAsJsonObject().get("Given").getAsString() + " " + info.getAsJsonObject().get("Family").getAsString());
            user.addProperty("memberSince", info.getAsJsonObject().get("Created").getAsString());
            names.add(user);
        });
        return names;
    }

    @Override
    public JsonArray getUserIdByCouId(Integer couId, boolean admin) {
        Map<String, String> params = new HashMap<>();
        params.put("couid", couId.toString());
        if (admin) {
            params.put("admin", "true");
        }
        JsonElement response = httpUtils.get("identifiers.json", params);
        JsonArray infos = (response != null) ? response.getAsJsonObject().get("Identifiers").getAsJsonArray() : new JsonArray();
        JsonArray emails = new JsonArray();
        infos.forEach(info -> {
            JsonObject user = new JsonObject();
            user.addProperty("id", info.getAsJsonObject().get("Identifier").getAsString());
            user.addProperty("memberSince", info.getAsJsonObject().get("Created").getAsString());
            emails.add(user);
        });
        return emails;
    }

    @Override
    public void assignMemberRole(Integer coPersonId, Integer couId, Integer id) {
        if (id != null) {
            httpUtils.put("co_person_roles/" + id.toString() + ".json", jsonUtils.coPersonRoles(coPersonId, couId, "Active"));
        } else {
            httpUtils.post("co_person_roles.json", jsonUtils.coPersonRoles(coPersonId, couId, "Active"));
        }
    }

    @Override
    public void removeMemberRole(Integer coPersonId, Integer couId, Integer id) {
        if (id != null) {
            httpUtils.put("co_person_roles/" + id.toString() + ".json", jsonUtils.coPersonRoles(coPersonId, couId, "Deleted"));
        }
    }

    @Override
    public Integer createRole(Role role) {
        JsonElement element = httpUtils.post("cous.json", jsonUtils.createNewCou(role));
        return element.getAsJsonObject().get("Id").getAsInt();
    }

    @Override
    public String getUserEmail(Integer coPersonId) {
        Map<String, String> params = new HashMap<>();
        params.put("copersonid", coPersonId.toString());
        JsonElement response = httpUtils.get("email_addresses.json", params);
        JsonObject info = (response != null) ? response.getAsJsonObject().get("EmailAddresses").getAsJsonArray().get(0).getAsJsonObject() : null;
        return (info != null) ? info.getAsJsonObject().get("Mail").getAsString() : null;
    }

    @Override
    public String getUserNames(Integer coPersonId) {
        Map<String, String> params = new HashMap<>();
        params.put("copersonid", coPersonId.toString());
        JsonElement response = httpUtils.get("names.json", params);
        JsonObject info = (response != null) ? response.getAsJsonObject().get("Names").getAsJsonArray().get(0).getAsJsonObject() : null;
        return (info != null) ? info.getAsJsonObject().get("Given").getAsString() + " " + info.getAsJsonObject().get("Family").getAsString() : null;
    }

    @Override
    public String getUserId(Integer coPersonId) {
        Map<String, String> params = new HashMap<>();
        params.put("copersonid", coPersonId.toString());
        JsonElement response = httpUtils.get("identifiers.json", params);
        JsonObject info = (response != null) ? response.getAsJsonObject().get("Identifiers").getAsJsonArray().get(0).getAsJsonObject() : null;
        return (info != null) ? info.getAsJsonObject().get("Identifier").getAsString() : null;
    }

    @Override
    public void assignAdminRole(Integer coPersonId, Integer couId) {
        JsonObject group = getCouAdminGroup(couId);
        if (group != null) {
            httpUtils.post("co_group_members.json", jsonUtils.coGroupMembers(group.get("Id").getAsInt(), coPersonId, true));
        }
    }

    @Override
    public void removeAdminRole(Integer coPersonId, Integer couId) {
        JsonObject adminGroup = this.getCouAdminGroup(couId);
        JsonArray admins = this.getGroupMembers(adminGroup.get("Id").getAsInt());
        Integer id = null;
        for (JsonElement admin : admins) {
            if (admin.getAsJsonObject().get("Person").getAsJsonObject().get("Id").getAsInt() == coPersonId) {
                id = admin.getAsJsonObject().get("Id").getAsInt();
            }
        }
        if (id != null) {
            httpUtils.delete("co_group_members/" + id.toString() + ".json");
        }
    }

    @Override
    public Map<Integer, String> getCouNames(List<Integer> couIds) {
        Map<Integer, String> idNameMap = new HashMap<>();
        for (Integer id : couIds) {
            idNameMap.put(id, null);
        }

        JsonArray cous = getCous();
        int count = 0;
        int total = couIds.size();
        for (JsonElement cou : cous) {
            if (count < total) {
                if (idNameMap.containsKey(cou.getAsJsonObject().get("Id").getAsInt())) {
                    idNameMap.put(cou.getAsJsonObject().get("Id").getAsInt(), cou.getAsJsonObject().get("Name").getAsString());
                    count++;
                }
            } else {
                break;
            }
        }
        return idNameMap;
    }
}
